from six import string_types
from pypif.obj.common.pio import Pio


class FileReference(Pio):
    """
    Information about a file.
    """

    def __init__(self, relative_path=None, mime_type=None, sha256=None, md5=None, tags=None, **kwargs):
        """
        Constructor.

        :param relative_path: String with the relative path (from the location of this file) of the file.
        :param mime_type: String with the mime type of the file.
        :param sha256: String with the SHA-256 hash of the file.
        :param md5: String with the MD5 hash of the file.
        :param tags: List of strings or numbers that are tags for this object.
        :param kwargs: Dictionary of fields that are not supported.
        """
        super(FileReference, self).__init__(tags=tags, **kwargs)
        self._relative_path = None
        self.relative_path = relative_path
        self._mime_type = None
        self.mime_type = mime_type
        self._sha256 = None
        self.sha256 = None
        self._md5 = None
        self.md5 = None

    @property
    def relative_path(self):
        return self._relative_path

    @relative_path.setter
    def relative_path(self, relative_path):
        self._validate_type('relative_path', relative_path, string_types)
        self._relative_path = relative_path

    @relative_path.deleter
    def relative_path(self):
        self._relative_path = None

    @property
    def mime_type(self):
        return self._mime_type

    @mime_type.setter
    def mime_type(self, mime_type):
        self._validate_type('mime_type', mime_type, string_types)
        self._mime_type = mime_type

    @mime_type.deleter
    def mime_type(self):
        self._mime_type = None

    @property
    def sha256(self):
        return self._sha256

    @sha256.setter
    def sha256(self, sha256):
        self._validate_type('sha256', sha256, string_types)
        self._sha256 = sha256

    @sha256.deleter
    def sha256(self):
        self._sha256 = None

    @property
    def md5(self):
        return self._md5

    @md5.setter
    def md5(self, md5):
        self._validate_type('md5', md5, string_types)
        self._md5 = md5

    @md5.deleter
    def md5(self):
        self._md5 = None



from pypif.stats.common.stats_wrapper import Stats
from pypif.stats.common.stats_wrapper import StatsWrapper
from pypif.stats.common.field_stats import FieldStats
from pypif.stats.common.property_stats_wrapper import PropertyStatsWrapper


class SystemStats(Stats):
    """
    Class to store stats of a single system.
    """

    def __init__(self, count=None, names=None, chemical_formula=None, properties=None):
        """
        Constructor.

        :param count: Number of systems of this type.
        :param names: Dictionary or :class:`.FieldStats` object with stats of the system names.
        :param chemical_formula: Dictionary or :class:`.FieldStats` object with stats of the system chemical formula.
        :param properties: Dictionary or :class:`.PropertyStatsWrapper` object with stats of the system properties.
        """
        super(SystemStats, self).__init__(count=count)
        self.names = self._get_object(FieldStats, names)
        self.chemical_formula = self._get_object(FieldStats, chemical_formula)
        self.properties = self._get_object(PropertyStatsWrapper, properties)


class SystemStatsWrapper(StatsWrapper):
    """
    Class to store stats of systems.
    """

    def __init__(self, count=None, common=None):
        """
        Constructor.

        :param count: Number of systems.
        :param common: Dictionary or :class:`.SystemStats` object with the stats of the system.
        """
        super(SystemStatsWrapper, self).__init__(count=count)
        self.common = self._get_object(SystemStats, common)



from pypif.stats.common.stats_wrapper import Stats
from pypif.stats.common.stats_wrapper import StatsWrapper
from pypif.stats.common.field_stats import FieldStats


class PropertyStats(Stats):
    """
    Class to store stats about a single property.
    """

    def __init__(self, count=None, name=None, value=None, units=None):
        """
        Constructor.

        :param count: Count for this object.
        :param name: Dictionary or :class:`.FieldStats` object with stats of the property name.
        :param value: Dictionary or :class:`.FieldStats` object with stats of the property value.
        :param units: Dictionary or :class:`.FieldStats` object with stats of the property units.
        """
        super(PropertyStats, self).__init__(count=count)
        self.name = self._get_object(FieldStats, name)
        self.value = self._get_object(FieldStats, value)
        self.units = self._get_object(FieldStats, units)


class PropertyStatsWrapper(StatsWrapper):
    """
    Class to store stats about a set of properties.
    """

    def __init__(self, count=None, common=None):
        """
        Constructor.

        :param count: Count of properties.
        :param common: List of Dictionary or :class:`.PropertyStats` objects with stats of the properties.
        """
        super(PropertyStatsWrapper, self).__init__(count=count)
        self.common = self._get_object(PropertyStats, common)



from pypif.stats.common.stats_wrapper import Stats


class TermAndCount(Stats):
    """
    Class to store information about a single term.
    """

    def __init__(self, count=None, term=None):
        """
        Information about a single term and its count.

        :param count: Total count for this term.
        :param term: String with the term to save.
        """
        super(TermAndCount, self).__init__(count=count)
        self.term = term


class FieldStats(Stats):
    """
    Class to store stats for a single field.
    """

    def __init__(self, count=None, numeric_count=None, min=None, max=None, terms=None):
        """
        Constructor.

        :param count: Total count for this field.
        :param numeric_count: Number of numeric values for this field.
        :param min: Minimum of the numeric values for this field.
        :param max: Maximum of the numeric values for this field.
        :param terms: List of dictionaries or :class:`.TermAndCount` objects with stats about terms.
        """
        super(FieldStats, self).__init__(count=count)
        self.numeric_count = numeric_count
        self.min = min
        self.max = max
        self.terms = self._get_object(TermAndCount, terms)



from pypif.obj.common.property import Property
from pypif.obj.common.quantity import Quantity
from pypif.obj.common.display_item import DisplayItem 
from pypif.obj.common.id import Id
from pypif.obj.common.instrument import Instrument
from pypif.obj.common.license import License
from pypif.obj.common.method import Method
from pypif.obj.common.software import Software
from pypif.obj.common.source import Source
from pypif.obj.common.value import Value
from pypif.obj.common.reference import Reference
from pypif.obj import System as Subsystem
from pypif.pif import dumps


def new_keypair(key, value, ambig, unambig):
    """
    Check new keypair against existing unambiguous dict

    :param key: of pair
    :param value: of pair
    :param ambig: set of keys with ambig decoding
    :param unambig: set of keys with unambig decoding
    :return:
    """
    if key in ambig:
        return

    if key in unambig and value != unambig[key]:
            ambig.add(key)
            del unambig[key]
            return

    unambig[key] = value
    return


def add_child_ambig(child_ambig, child_unambig, ambig, unambig):
    """
    Add information about decodings of a child object

    :param child_ambig: ambiguous set from child
    :param child_unambig: unambiguous set from child
    :param ambig: set of keys storing ambig decodings
    :param unambig: dictionary storing unambiguous decodings
    :return:
    """
    for k in child_ambig:
        ambig.add(k)
        del unambig[k]

    for k, v in child_unambig.items():
        new_keypair(k, v, ambig, unambig)

    return


class ReadView():

    def __init__(self, system):
        """
        Setup a ReadView by recursing through the pif objects

        :param system: to process into a ReadView; doesn't need to be a system
        """

        self.raw = system

        # List of classes to generate an index from
        self.inlines = [
            (Subsystem, "uid"),
            (Property, "name"),
            (Quantity, "name"),
            (Id, "name"),
            (Instrument, "name"),
            (License, "name"),
            (Method, "name"),
            (Value, "name"),
            (Software, "name"),
            (Source, "producer"),
            (DisplayItem, "title"),
            (Reference, "doi"),
        ]

        # dictionary of unambiguous short keys
        self.unambig = {}
        # set of keys where there is ambiguity
        self.ambig = set()

        # iterate over the members of the system
        for k in system.__dict__:
            added = False

            # check if list
            if isinstance(getattr(system, k), list):
                for t, name in self.inlines:
                    # check if we know how to decode the type
                    if isinstance(getattr(system, k)[0], t):
                        # populate a dictionary
                        parsed = {}
                        for x in getattr(system, k):
                            # define a key
                            new_key = getattr(x, name)
                            new_val = ReadView(x)
                            parsed[new_key] = new_val
                            new_keypair(new_key, new_val, self.ambig, self.unambig)
                            add_child_ambig(new_val.ambig, new_val.unambig, self.ambig, self.unambig)
                        # set the property in the readview to point to the dictionary
                        setattr(self, k[1:], parsed)
                        added = True
                        break
                if not added:
                    # just mirror the property
                    setattr(self, k[1:], getattr(system, k))
            else:
                # Same logic as above, but not wrapped in a list
                for t, name in self.inlines:
                    if isinstance(getattr(system, k), t):
                        new_key = getattr(getattr(system, k), name)
                        new_val = ReadView(getattr(system, k))
                        new_keypair(new_key, new_val, self.ambig, self.unambig)
                        add_child_ambig(new_val.ambig, new_val.unambig, self.ambig, self.unambig)
                        setattr(self, k[1:], {new_key: new_val})
                        added = True
                        break
                if not added:
                    setattr(self, k[1:], getattr(system, k))

    def keys(self):
        return self.unambig.keys()

    def __getitem__(self, key):
        if key not in self.unambig:
            raise KeyError(key + " not defined unambiguously")
        return self.unambig[key]

    def __ne__(self, other):
        if not isinstance(other, ReadView):
            return True

        return dumps(other.raw) != dumps(self.raw)

    def __eq__(self, other):
        return not self.__ne__(other)


from pypif.util.serializable import Serializable


class Stats(Serializable):
    """
    Class to store stats for a single object.
    """

    def __init__(self, count=None):
        """
        Constructor.

        :param count: Integer with the count for this object.
        """
        self.count = count


class StatsWrapper(Serializable):
    """
    Class to store stats for a set of objects.
    """

    def __init__(self, count):
        """
        Constructor.

        :param count: Integer with the count for this object.
        """
        self.count = count
